// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <libxml/parser.h>

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "globals.hh"
#include "Iterator.hh"
#include "stringAux.hh"
#include "CharMapper.hh"
#include "FontManager.hh"
#include "EntitiesTable.hh"
#include "MathMLParseFile.hh"

// CharMapper: this class is used to map a Unicode char and a FontAttributes structures
// into a native char index and a native font to draw that char.
// Information on the native fonts available in the system and the actual mapping
// of the glyphs inside them is given externally by means of a configuration XML file.

// some auxiliary functions defined below

static unsigned deMask(unsigned);
static Char parseCode(xmlNode*);

// CharMapper, default constructo
CharMapper::CharMapper()
{
}

CharMapper::~CharMapper()
{
  for (Iterator<FontMap*> i(content); i.More(); i.Next()) {
    assert(i() != NULL);
    delete i();
  }
}

// FontifyChar: this is the main method for this class. It takes a Unicode
// char ch and a FontAttributes structures and tries to fontify the character, i.e.
// it finds a font whose attributes are as close as possible to FA and which is
// capable of render CH. Moreover, it returns a mapping function, i.e. a conversion
// function which takes CH and return an index to the corresponding glyph within
// the selected font.
FontifiedChar*
CharMapper::FontifyChar(const FontAttributes& fa, Char ch, bool stretchy) const
{
  unsigned mask = fa.mask;
  unsigned oldMask = mask;

  // the search proceeds in a top-down manner, that is, fonts listed
  // first within the config file are tested first. These should be the
  // most specific and system-dependent fonts which allow, when present,
  // the best possible rendering.
  do {
    for (Iterator<FontMap*> fontMap(content); fontMap.More(); fontMap.Next()) {
      assert(fontMap() != NULL);
      FontifiedChar* fMap = fontMap()->FontifyChar(mask, fa, ch, stretchy);
      if (fMap != NULL) return fMap;
    }

    // oldMask is needed because we want to make at least one attempt with
    // the mask set to 0.
    oldMask = mask;
    mask = deMask(mask);
  } while (oldMask != mask);

  // if the mask has reached 0 and we are still here, then there is no font available

  return NULL;
}

StretchId
CharMapper::GetStretch(Char ch) const
{
  for (Iterator<FontMap*> fontMap(content); fontMap.More(); fontMap.Next()) {
    assert(fontMap() != NULL);
    StretchId res = fontMap()->GetStretch(ch);
    if (res != STRETCH_NO) return res;
  }

  return STRETCH_NO;
}

StretchId
CharMapper::FontMap::GetStretch(Char ch) const
{
  for (Iterator<CharMap*> charMap(single[CHAR_HASH(ch)]); charMap.More(); charMap.Next()) {
    assert(charMap() != NULL);
    if (charMap()->MapsStretchyChar(ch)) {
      if (charMap()->stretchy.direction != STRETCH_NO) return charMap()->stretchy.direction;
    }
  }

  return STRETCH_NO;
}

FontifiedChar*
CharMapper::FontMap::FontifyChar(unsigned mask, const FontAttributes& fa, Char ch, bool stretchy) const
{
  if (!attributes.CompatibleWith(mask, fa)) return NULL;

  const CharMap* m = NULL;

  if (!stretchy) {
    for (Iterator<CharMap*> charMap(multi); m == NULL && charMap.More(); charMap.Next()) {
      assert(charMap() != NULL);
      if (charMap()->MapsChar(ch)) m = charMap();
    }
  }

  if (m == NULL) {
    unsigned idx = ch & CHAR_MAP_HASH_MASK;
    for (Iterator<CharMap*> charMap(single[idx]); m == NULL && charMap.More(); charMap.Next()) {
      assert(charMap() != NULL);
      if ((!stretchy && charMap()->MapsSimpleChar(ch)) ||
	  (stretchy && charMap()->MapsStretchyChar(ch))) m = charMap();
    }
  }

  if (m == NULL) return NULL;

  FontAttributes fReal(mask, fa, attributes);
  const Font* font = fontManager.GetFont(fReal);
  if (font == NULL) return NULL;

  FontifiedChar* fChar = new FontifiedChar;
  fChar->font    = font;
  fChar->nch     = m->Map(ch);
  fChar->charMap = m;
  return fChar;
}

// Load: loads a configuration file
bool
CharMapper::Load(const char* fileName)
{
  assert(fileName != NULL);

  xmlDocPtr doc = MathMLParseFile(fileName, false);
  if (doc == NULL) return false;

  ParseFontMapList(xmlDocGetRootElement(doc));
  xmlFreeDoc(doc);

  return true;
}

void
CharMapper::ParseFontMapList(xmlNodePtr node)
{
  if (node == NULL) return;
  if (node->name == NULL) return;
  // a conf file is made of a single <unimap> element
  if (xmlStrcmp(node->name, XMLSTRING("unimap"))) return;

  xmlNodePtr p = node->children;
  while (p != NULL) {
    // every child of <unimap> must be a particular font map
    ParseFontMap(p);
    p = p->next;
  }
}

void
CharMapper::ParseFontMap(xmlNodePtr node)
{
  if (node == NULL) return;
  if (node->name == NULL) return;
  if (xmlStrcmp(node->name, XMLSTRING("font"))) return;

  FontMap* fontMap = new FontMap;
  fontMap->attributes.mask = 0;

  xmlChar* value = NULL;

  value = xmlGetProp(node, XMLSTRING("foundry"));
  if (value != NULL) {
    fontMap->attributes.foundry = CSTRING(xmlStrdup(value));
    fontMap->attributes.mask |= FONT_HAS_FOUNDRY;
  }

  value = xmlGetProp(node, XMLSTRING("family"));
  if (value != NULL) {
    fontMap->attributes.family = CSTRING(xmlStrdup(value));
    fontMap->attributes.mask |= FONT_HAS_FAMILY;
  }
  
  value = xmlGetProp(node, XMLSTRING("style"));
  if (value != NULL) {
    if (!xmlStrcmp(value, XMLSTRING("normal"))) {
      fontMap->attributes.style = FONT_STYLE_NORMAL;
      fontMap->attributes.mask |= FONT_HAS_STYLE;
    } else if (!xmlStrcmp(value, XMLSTRING("italic"))) {
      fontMap->attributes.style = FONT_STYLE_ITALIC;
      fontMap->attributes.mask |= FONT_HAS_STYLE;
    }
  }

  value = xmlGetProp(node, XMLSTRING("weight"));
  if (value != NULL) {
    if (!xmlStrcmp(value, XMLSTRING("normal"))) {
      fontMap->attributes.weight = FONT_WEIGHT_NORMAL;
      fontMap->attributes.mask |= FONT_HAS_WEIGHT;
    } else if (!xmlStrcmp(value, XMLSTRING("bold"))) {
      fontMap->attributes.weight = FONT_WEIGHT_BOLD;
      fontMap->attributes.mask |= FONT_HAS_WEIGHT;
    }
  }

  value = xmlGetProp(node, XMLSTRING("size"));
  if (value != NULL) {
    char* suffix;
    float v = strtod(CSTRING(value), &suffix);
    UnitId unitId = UNIT_PT;
    if (suffix != NULL) {
      if (!strcmp(suffix, "px")) unitId = UNIT_PX;
    }
    fontMap->attributes.size.Set(v, unitId);
    fontMap->attributes.mask |= FONT_HAS_SIZE;
  }

  value = xmlGetProp(node, XMLSTRING("charset"));
  if (value != NULL) {
    fontMap->attributes.charSet = CSTRING(xmlStrdup(value));
    fontMap->attributes.mask |= FONT_HAS_CHARSET;
  }

  xmlNodePtr p = node->children;
  while (p != NULL) {
    ParseCharMap(p, fontMap);
    p = p->next;
  }

  content.Append(fontMap);
}

void
CharMapper::ParseCharMap(xmlNodePtr node, FontMap* fontMap)
{
  assert(fontMap != NULL);

  if (node == NULL) return;
  if (node->name == NULL) return;
  if      (!xmlStrcmp(node->name, XMLSTRING("range"))) ParseRange(node, fontMap);
  else if (!xmlStrcmp(node->name, XMLSTRING("multi"))) ParseMulti(node, fontMap);
  else if (!xmlStrcmp(node->name, XMLSTRING("single"))) ParseSingle(node, fontMap);
  else if (!xmlStrcmp(node->name, XMLSTRING("stretchy"))) ParseStretchy(node, fontMap);
}

void
CharMapper::ParseRange(xmlNodePtr node, FontMap* fontMap)
{
  CharMap* charMap = new CharMap;
  charMap->type = CHAR_MAP_RANGE;

  xmlChar* value = xmlGetProp(node, XMLSTRING("first"));
  if (value == NULL) {
    delete charMap;
    return;
  }
  charMap->range.first = strtoul(CSTRING(value), NULL, 0);

  value = xmlGetProp(node, XMLSTRING("last"));
  if (value == NULL) {
    delete charMap;
    return;
  }
  charMap->range.last = strtoul(CSTRING(value), NULL, 0);

  value = xmlGetProp(node, XMLSTRING("offset"));
  if (value == NULL) {
    delete charMap;
    return;
  }
  charMap->range.offset = strtoul(CSTRING(value), NULL, 0);

  if (charMap->range.last < charMap->range.first) {
    delete charMap;
    return;
  }

  fontMap->multi.Append(charMap);
}

void
CharMapper::ParseMulti(xmlNodePtr node, FontMap* fontMap)
{
  CharMap* charMap = new CharMap;
  charMap->type = CHAR_MAP_MULTI;

  xmlChar* value = xmlGetProp(node, XMLSTRING("first"));
  if (value == NULL) {
    delete charMap;
    return;
  }
  charMap->multi.first = strtoul(CSTRING(value), NULL, 0);

  value = xmlGetProp(node, XMLSTRING("last"));
  if (value == NULL) {
    delete charMap;
    return;
  }
  charMap->multi.last = strtoul(CSTRING(value), NULL, 0);

  if (charMap->multi.last < charMap->multi.first) {
    delete charMap;
    return;
  }

  value = xmlGetProp(node, XMLSTRING("index"));
  if (value == NULL) {
    delete charMap;
    return;
  }
  charMap->multi.index = new NativeChar[charMap->multi.last - charMap->multi.first + 1];

  const char* ptr = CSTRING(value);
  for (NativeChar ch = charMap->multi.first; ch < charMap->multi.last; ch++) {
    char* newPtr;
    charMap->multi.index[ch - charMap->multi.first] = strtoul(ptr, &newPtr, 0);
    ptr = newPtr;
  }

  fontMap->multi.Append(charMap);
}

void
CharMapper::ParseSingle(xmlNodePtr node, FontMap* fontMap)
{
  CharMap* charMap = new CharMap;
  charMap->type = CHAR_MAP_SINGLE;

  charMap->single.code = parseCode(node);
  if (charMap->single.code == 0) {
    delete charMap;
    return;
  }

  xmlChar* value = xmlGetProp(node, XMLSTRING("index"));
  if (value == NULL) {
    delete charMap;
    return;
  }
  charMap->single.index = strtoul(CSTRING(value), NULL, 0);

  fontMap->single[CHAR_HASH(charMap->single.code)].Append(charMap);
}

void
CharMapper::ParseStretchy(xmlNodePtr node, FontMap* fontMap)
{
  CharMap* charMap = new CharMap;
  charMap->type = CHAR_MAP_STRETCHY;

  for (unsigned i = 0; i < MAX_SIMPLE_CHARS; i++) charMap->stretchy.simple[i] = 0;
  for (unsigned i = 0; i < SC_REPEAT + 1; i++) charMap->stretchy.compound[i] = 0;

  charMap->stretchy.code = parseCode(node);
  if (charMap->stretchy.code == 0) {
    delete charMap;
    return;
  }

  xmlChar* value = xmlGetProp(node, XMLSTRING("direction"));
  if (value != NULL)
    if (!xmlStrcmp(value, XMLSTRING("horizontal")))
      charMap->stretchy.direction = STRETCH_HORIZONTAL;
    else if (!xmlStrcmp(value, XMLSTRING("vertical")))
      charMap->stretchy.direction = STRETCH_VERTICAL;
    else
      charMap->stretchy.direction = STRETCH_NO;
  else
    charMap->stretchy.direction = STRETCH_NO;

  xmlNodePtr p = node->children;
  while (p != NULL) {
    if (!xmlStrcmp(p->name, XMLSTRING("simple"))) ParseStretchySimple(p, charMap);
    else if (!xmlStrcmp(p->name, XMLSTRING("compound"))) ParseStretchyCompound(p, charMap);

    p = p->next;
  }

  fontMap->single[CHAR_HASH(charMap->stretchy.code)].Append(charMap);
}

void
CharMapper::ParseStretchySimple(xmlNodePtr node, CharMap* charMap)
{
  assert(node != NULL);
  assert(charMap != NULL);

  xmlChar* value = xmlGetProp(node, XMLSTRING("index"));
  if (value == NULL) return;

  const char* ptr = CSTRING(value);
  for (unsigned i = 0; i < MAX_SIMPLE_CHARS && ptr != NULL && *ptr != '\0'; i++) {
    char* newPtr;
    if (i < 4) charMap->stretchy.simple[i] = strtoul(ptr, &newPtr, 0);
    ptr = newPtr;
  }
}

void
CharMapper::ParseStretchyCompound(xmlNodePtr node, CharMap* charMap)
{
  assert(node != NULL);
  assert(charMap != NULL);

  xmlChar* value = xmlGetProp(node, XMLSTRING("index"));
  if (value == NULL) return;

  const char* ptr = CSTRING(value);
  for (unsigned i = 0; i < SC_REPEAT + 1 && ptr != NULL && *ptr != '\0'; i++) {
    char* newPtr;
    if (i < SC_REPEAT + 1) charMap->stretchy.compound[i] = strtoul(ptr, &newPtr, 0);
    ptr = newPtr;
  }
}

/////////////////////////
// auxiliary functions //
/////////////////////////

static unsigned
deMask(unsigned mask)
{
  if ((mask & FONT_HAS_CHARSET) != 0) mask &= ~FONT_HAS_CHARSET;
  else if ((mask & FONT_HAS_FAMILY) != 0) mask &= ~FONT_HAS_FAMILY;
  else if ((mask & FONT_HAS_STYLE) != 0) mask &= ~FONT_HAS_STYLE;
  else if ((mask & FONT_HAS_WEIGHT) != 0) mask &= ~FONT_HAS_WEIGHT;
  else if ((mask & FONT_HAS_SIZE) != 0) mask &= ~FONT_HAS_SIZE;

  return mask;
}

static Char
parseCode(xmlNode* node)
{
  assert(node != NULL);

  xmlChar* value = xmlGetProp(node, XMLSTRING("code"));
  if (value != NULL) {
    if (*value == '\0') return 0;

    if (*value == '0' && tolower(*(value + 1)) == 'x') return strtoul(CSTRING(value), NULL, 0);
    else if (isPlain(*value) && *(value + 1) == '\0') return *value;

    if (verbose)
      fprintf(stderr, "WARNING: UTF8 character(s) inside font configuration file (ignored)\n");

    return 0;
  } else {
    value = xmlGetProp(node, XMLSTRING("name"));
    if (value == NULL) return 0;

    String* s = entitiesTable.GetEntityContent(value);

    Char ch = 0;

    if (s != NULL && s->GetLength() > 0) {
      assert(s != NULL);
      ch = s->GetChar(0);
      delete s;
    } else if (s == NULL) {
      if (verbose)
	fprintf(stderr, "WARNING: unknown entity `%s' in font configuration file (ignored)\n", value);
    }

    return ch;
  }
}
