// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifndef MathMLElement_hh
#define MathMLElement_hh

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "keyword.hh"
#include "MathMLFrame.hh"
#include "BoundingBox.hh"
#include "DrawingArea.hh"
#include "AttributeSignature.hh"
#include "MathMLAttributeList.hh"

// MathMLElement: base class for every MathML Element
class MathMLElement: public MathMLFrame {
public:
  MathMLElement(TagId tag = TAG_NOTVALID);
  virtual const AttributeSignature* GetAttributeSignature(AttributeId) const;
  virtual void Dump(void) const;      // for debugging purposes
  virtual void Normalize(void);       // normalization
  virtual void Setup(class RenderingEnvironment*); // setup attributes
  void         DoBoxedLayout(LayoutId id) { DoBoxedLayout(id, (id == LAYOUT_MAX) ? BREAK_NO : BREAK_GOOD); }
  void         DoBoxedLayout(LayoutId id, BreakId bid) { DoBoxedLayout(id, bid, 0); }
  void         DoBoxedLayout(LayoutId id, scaled w) { DoBoxedLayout(id, BREAK_GOOD, w); }
  virtual void DoBoxedLayout(LayoutId, BreakId, scaled);
  virtual void DoLayout(LayoutId, class Layout&);   // layout as breakable
  virtual void DoStretchyLayout(void);          // layout stretchy components
  virtual void SetPosition(scaled, scaled);
  virtual void Freeze(void);
  virtual void RenderBackground(const DrawingArea&);
  virtual void Render(const DrawingArea&);
  virtual void SetDirty(const Rectangle* = NULL);
  virtual MathMLElement* Inside(scaled, scaled);
  virtual bool IsElement(void) const;
  virtual ~MathMLElement();

  GraphicsContext* GetGraphicsContext(void) const { return gc; }

  // attributes
  const String* GetAttributeDefault(AttributeId) const;
  const String* GetAttribute(AttributeId,
			     const RenderingEnvironment* = NULL,
			     bool = true) const;
  const Value* ParseAttribute(AttributeId, const String*) const;
  const Value* GetDefaultAttributeValue(AttributeId) const;
  const Value* GetAttributeValue(AttributeId,
				 const RenderingEnvironment* = NULL,
				 bool = true) const;
  const Value* ResolveAttributeValue(AttributeId,
				     const RenderingEnvironment* = NULL) const;
  static const Value* Resolve(const Value*,
			      const RenderingEnvironment*,
			      int = -1, int = -1);
  bool IsSet(AttributeId) const;

  // bounding boxes
  const BoundingBox& GetMinBoundingBox(void) const { return minBox; }
  const BoundingBox& GetMaxBoundingBox(void) const { return maxBox; }
  virtual void   GetLinearBoundingBox(BoundingBox&) const;
  virtual void   RecalcBoundingBox(LayoutId, scaled = 0);

  // some queries
  TagId        	 IsA(void) const { return tag; }
  virtual bool 	 IsSpaceLike(void) const;
  virtual bool 	 IsExpanding(void) const;
  virtual bool 	 IsInside(scaled, scaled) const;
  bool         	 HasLayout(void) const { return layout != NULL; }
  bool         	 IsShaped(void) const { return shape != NULL; }
  RGBValue     	 GetParentBackgroundColor(void) const;
  unsigned     	 GetDepth(void) const;
  const Layout&  GetLayout(void) const;
  const Shape&   GetShape(void) const;
  virtual scaled GetLeftEdge(void) const;
  virtual scaled GetRightEdge(void) const;

  MathMLAttributeList attributes;

#ifdef MATHML_STATS
  static unsigned GetCounter(void) { return elementCounter; }
#endif

protected:
  void ResetLayout(void);
  const AttributeSignature* GetAttributeSignatureAux(AttributeId,
						     AttributeSignature[]) const;
  void ConfirmLayout(LayoutId);

  BoundingBox minBox;
  BoundingBox maxBox;
  class Layout* layout;
  const class Shape* shape;
  class GraphicsContext* gc;

private:
  TagId tag;

#ifdef MATHML_STATS
  static unsigned elementCounter;
#endif
};

typedef MathMLElement* MathMLElementPtr;

#define TO_ELEMENT(object) (dynamic_cast<MathMLElement*>(object))

inline void
MathMLElement::ConfirmLayout(LayoutId id)
{
  switch (id) {
  case LAYOUT_AUTO: break;
  case LAYOUT_MIN: minBox = box; break;
  case LAYOUT_MAX: maxBox = box; break;
  }
}

#endif
