// Copyright (C) 2000, Luca Padovani <luca.padovani@cs.unibo.it>.
// 
// This file is part of GtkMathView, a Gtk widget for MathML.
// 
// GtkMathView is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// GtkMathView is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with GtkMathView; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
// For details, see the GtkMathView World-Wide-Web page,
// http://cs.unibo.it/~lpadovan/mml-widget, or send a mail to
// <luca.padovani@cs.unibo.it>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <stddef.h>

#include "Iterator.hh"
#include "StringUnicode.hh"
#include "MathMLTableElement.hh"
#include "MathMLTableCellElement.hh"

MathMLTableElement::MathMLTableElement(): MathMLContainerElement(TAG_MTABLE)
{
  nRows    = 0;
  nColumns = 0;
  cell     = NULL;
  column   = NULL;
  row      = NULL;
}

const AttributeSignature*
MathMLTableElement::GetAttributeSignature(AttributeId id) const
{
  static AttributeSignature sig[] = {
    { ATTR_ALIGN,           tableAlignParser,         new StringC("axis") },
    { ATTR_ROWALIGN,        rowAlignListParser,       new StringC("baseline") },
    { ATTR_COLUMNALIGN,     columnAlignListParser,    new StringC("center") },
    { ATTR_GROUPALIGN,      groupAlignListListParser, new StringC("{ left }") },
    { ATTR_ALIGNMENTSCOPE,  booleanListParser,        new StringC("true") },
    { ATTR_COLUMNWIDTH,     columnWidthListParser,    new StringC("auto") },
    { ATTR_ROWSPACING,      numberUnitListParser,     new StringC("1.0ex") },
    { ATTR_COLUMNSPACING,   spaceListParser,          new StringC("0.8em") },
    { ATTR_ROWLINES,        lineTypeListParser,       new StringC("none") },
    { ATTR_COLUMNLINES,     lineTypeListParser,       new StringC("none") },
    { ATTR_FRAME,           lineTypeParser,           new StringC("none") },
    { ATTR_FRAMESPACING,    tableFrameSpacingParser,  new StringC("0.4em 0.5ex") },
    { ATTR_EQUALROWS,       booleanParser,            new StringC("true") },
    { ATTR_EQUALCOLUMNS,    booleanParser,            new StringC("true") },
    { ATTR_DISPLAYSTYLE,    booleanParser,            new StringC("false") },
    { ATTR_SIDE,            tableSideParser,          new StringC("right") },
    { ATTR_MINLABELSPACING, numberUnitParser,         new StringC("0.8em") },
    { ATTR_WIDTH,           tableWidthParser,         new StringC("auto") },
    { ATTR_NOTVALID,        NULL,                     NULL }
  };

  const AttributeSignature* signature = GetAttributeSignatureAux(id, sig);
  if (signature == NULL) signature = MathMLContainerElement::GetAttributeSignature(id);

  return signature;
}

MathMLTableElement::~MathMLTableElement()
{
}

void
MathMLTableElement::Normalize()
{
  if (content.GetSize() == 0) {
    MathMLTableRowElement* mtr = new MathMLTableRowElement;
    mtr->SetParent(this);
    content.Append(mtr);

    MathMLTableCellElement* mtd = new MathMLTableCellElement;
    mtd->SetParent(mtr);
    mtr->content.Append(mtd);
  }

  for (unsigned i = 0; i < content.GetSize(); i++) {
    MathMLElement* elem = content.RemoveFirst();
    assert(elem != NULL);

    if (elem->IsA() != TAG_MTR && elem->IsA() != TAG_MLABELEDTR) {
      MathMLTableRowElement *inferredTableRow = new MathMLTableRowElement;
      inferredTableRow->content.Append(elem);

      elem->SetParent(inferredTableRow);
      inferredTableRow->SetParent(this);
      elem = inferredTableRow;
    }
    elem->Normalize();
    content.Append(elem);
  }
}

void
MathMLTableElement::SetPosition(scaled x, scaled y)
{
  position.x = x;
  position.y = y;

  scaled yOffset = frameVerticalSpacing - box.ascent;

  for (unsigned i = 0; i < nRows; i++) {
    scaled xOffset = frameHorizontalSpacing;

    if (row[i].mtr != NULL) row[i].mtr->SetPosition(x + xOffset, y + yOffset + row[i].ascent);

    for (unsigned j = 0; j < nColumns; j++) {
      TableCell* cell = GetCell(i, j);

      if (cell->mtd != NULL && !cell->spanned) {
	const BoundingBox& cellBox = cell->mtd->GetBoundingBox();
	cell->mtd->SetPosition(x + xOffset, y + yOffset + cellBox.ascent);
      }

      xOffset += column[j].width + column[j].spacing;
    }

    yOffset += row[i].GetHeight() + row[i].spacing;
  }
}

void
MathMLTableElement::RenderTableBackground(const DrawingArea& area)
{
  if (gc == NULL) {
    gc = area.CreateGraphicsContext();
    assert(gc != NULL);
  }

  if (IsSelected() || HasDirtyBackground()) {
    if (IsSelected()) gc->SetColor(SELECTED_COLOR);
    else if (HasDirtyBackground()) gc->SetColor(GetParentBackgroundColor());

    area.Clear(gc, position.x, position.y, box);

    return;
  }

#if 0   
  for (unsigned i = 0; i < nRows; i++) {
    assert(row[i].mtr != NULL);
    if (row[i].mtr->IsSelected()) {
      row[i].mtr->RenderBackground(area);
      return;
    }
  }
#endif
}

void
MathMLTableElement::Render(const DrawingArea& area)
{
  if (!HasDirtyChildren()) return;

  //RenderBackground(area);
  //assert(gc != NULL);

  MathMLContainerElement::Render(area);

  gc->SetColor(color);

  if (frame != TABLE_LINE_NONE) {
    if (frame == TABLE_LINE_DASHED)
      gc->SetLineStyle(LINE_STYLE_DASHED);
    else
      gc->SetLineStyle(LINE_STYLE_SOLID);
    area.DrawRectangle(gc, position.x, position.y, box);
  }

  scaled yOffset = frameVerticalSpacing - box.ascent;
  for (unsigned i = 0; i < nRows; i++) {
    scaled xOffset = frameHorizontalSpacing;
    for (unsigned j = 0; j < nColumns; j++) {
      TableCell* cell = GetCell(i, j);

      if (i != nRows - 1 && row[i].lineType != TABLE_LINE_NONE) {
	// horizontal lines
	if (cell->rowSpan <= 1) {
	  scaled lineX = position.x + xOffset;
	  scaled lineY = position.y + yOffset + row[i].GetHeight() + row[i].spacing / 2;
	  scaled len = column[j].width;

	  if (j == 0) {
	    lineX -= frameHorizontalSpacing;
	    len   += frameHorizontalSpacing;
	  } else {
	    lineX -= column[j - 1].spacing / 2;
	    len   += column[j - 1].spacing / 2;
	  }
	  if (j == nColumns - 1) {
	    len   += frameHorizontalSpacing;
	  } else {
	    len   += column[j].spacing / 2;
	  }

	  if (row[i].lineType == TABLE_LINE_DASHED)
	    gc->SetLineStyle(LINE_STYLE_DASHED);
	  else
	    gc->SetLineStyle(LINE_STYLE_SOLID);
	  area.DrawLine(gc, lineX, lineY, lineX + len, lineY);
	}
      }

      if (j != nColumns - 1 && column[j].lineType != TABLE_LINE_NONE) {
	// vertical lines
	if (cell->colSpan <= 1) {
	  scaled lineX = position.x + xOffset + column[j].width + column[j].spacing / 2;
	  scaled lineY = position.y + yOffset;
	  scaled len = row[i].GetHeight();

	  if (i == 0) {
	    lineY -= frameVerticalSpacing;
	    len   += frameVerticalSpacing;
	  } else {
	    lineY -= row[i - 1].spacing / 2;
	    len   += row[i - 1].spacing / 2;
	  }
	  if (i == nRows - 1) {
	    len   += frameVerticalSpacing;
	  } else {
	    len   += row[i].spacing / 2;
	  }

	  if (column[j].lineType == TABLE_LINE_DASHED)
	    gc->SetLineStyle(LINE_STYLE_DASHED);
	  else
	    gc->SetLineStyle(LINE_STYLE_SOLID);
	  area.DrawLine(gc, lineX, lineY, lineX, lineY + len);
	}
      }

      xOffset += column[j].width + column[j].spacing;
    }

    yOffset += row[i].GetHeight() + row[i].spacing;
  }

  ResetDirty();
}

MathMLElement*
MathMLTableElement::Inside(scaled x, scaled y)
{
  if (!IsInside(x, y)) return NULL;
  
  for (unsigned i = 0; i < nRows; i++)
    for (unsigned j = 0; j < nColumns; j++)
      if (cell[i][j].mtd != NULL && !cell[i][j].spanned) {
	MathMLElement* inside = cell[i][j].mtd->Inside(x, y);
	
	if (inside != NULL) return inside;
      }

  return MathMLContainerElement::Inside(x, y);
}

bool
MathMLTableElement::IsExpanding() const
{
  return nFit > 0;
}

bool
MathMLTableElement::IsLabeled() const
{
  for (Iterator<MathMLElement*> p(content); p.More(); p.Next()) {
    assert(p() != NULL);
    if (p()->IsA() == TAG_MLABELEDTR) return true;
  }
  return false;
}

BreakId
MathMLTableElement::GetBreakability() const
{
  return BREAK_AUTO;
}
